/* GDM - The GNOME Display Manager
 * Copyright (C) 1998, 1999, 2000 Martin K. Petersen <mkp@mkp.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <string.h>

#include <fontconfig/fontconfig.h>

#include "gdm.h"
#include "gdmwm.h"
#include "gdmcommon.h"
#include "gdmconfig.h"
#include "gdmlanguages.h"
#include "greeter.h"
#include "greeter_configuration.h"
#include "greeter_item_pam.h"
#include "greeter_action_language.h"

#define LAST_LANGUAGE "Last"
#define DEFAULT_LANGUAGE "Default"

#ifndef I18N_CONFIG_FILE
#define I18N_CONFIG_FILE "/etc/sysconfig/i18n"
#endif

enum {
  LOCALE_COLUMN,
  TRANSLATED_NAME_COLUMN,
  UNTRANSLATED_NAME_COLUMN,
  IS_PREFERRED_COLUMN,
  NUM_COLUMNS
};

static GtkTreeModelFilter *filter_model = NULL;
static GtkListStore *lang_model = NULL;
static GtkWidget *dialog = NULL;
static gint savelang = GTK_RESPONSE_NO;
static gchar *current_language = NULL;
static gchar *dialog_selected_language = NULL;
static gboolean show_all_languages = FALSE;

typedef enum
{
  LOCALE_UP_TO_LANGUAGE = 0,
  LOCALE_UP_TO_COUNTRY,
  LOCALE_UP_TO_ENCODING,
  LOCALE_UP_TO_MODIFIER,
} LocaleScope;

static char *
get_less_specific_locale (const char *locale,
                          LocaleScope scope)
{
  char *generalized_locale;
  char *end;

  generalized_locale = strdup (locale);

  end = strchr (generalized_locale, '_');

  if (end != NULL && scope <= LOCALE_UP_TO_LANGUAGE)
    {
      *end = '\0';
      return generalized_locale;
    }

  end = strchr (generalized_locale, '.');

  if (end != NULL && scope <= LOCALE_UP_TO_COUNTRY)
    {
      *end = '\0';
      return generalized_locale;
    }

  end = strchr (generalized_locale, '@');

  if (end != NULL && scope <= LOCALE_UP_TO_ENCODING)
    {
      *end = '\0';
      return generalized_locale;
    }

  return generalized_locale;
}

static gboolean
locale_is_displayable (const char *locale)
{
  char *language_code;
  gboolean is_preferred;

  FcPattern *pattern;
  FcObjectSet *object_set;
  FcFontSet *font_set;

  is_preferred = FALSE;
  pattern = NULL;
  object_set = NULL;
  font_set = NULL;

  language_code = get_less_specific_locale (locale, LOCALE_UP_TO_LANGUAGE);

  pattern = FcPatternBuild (NULL, FC_LANG, FcTypeString, language_code, NULL);

  if (pattern == NULL)
    goto done;

  object_set = FcObjectSetBuild (NULL, NULL);

  if (object_set == NULL)
    goto done;

  font_set = FcFontList (NULL, pattern, object_set);

  if (font_set == NULL)
    goto done;

  is_preferred = (font_set->nfont > 0);

done:

  if (font_set != NULL)
    FcFontSetDestroy (font_set);

  if (object_set != NULL)
    FcObjectSetDestroy (object_set);

  if (pattern != NULL)
    FcPatternDestroy (pattern);

  g_free (language_code);
  return is_preferred;
}

static const gchar * const *
get_preferred_languages (void)
{
  GPtrArray *languages;
  const GList *locale_list, *tmp;
  char *locale_without_encoding, *i18n_file_contents;
  gsize i18n_file_length;
  const gchar *p, *q;
  GString *line;
  gsize i; 
  gboolean should_fallback = TRUE;
  static const gchar * const * preferred_languages;

  if (preferred_languages != NULL)
    return preferred_languages;

  languages = g_ptr_array_new ();

  locale_list = ve_i18n_get_language_list ("LC_MESSAGES");

  tmp = locale_list;
  while (tmp != NULL) {
        locale_without_encoding = 
            get_less_specific_locale (tmp->data, LOCALE_UP_TO_COUNTRY);
        g_ptr_array_add (languages, locale_without_encoding);
        tmp = tmp->next;
  }

  if (!g_file_get_contents (I18N_CONFIG_FILE, &i18n_file_contents,  
                            &i18n_file_length, NULL)) {
fallback: should_fallback = FALSE;
        i18n_file_contents = g_strdup ("SUPPORTED=en_US:de_DE:es_ES:fr_FR:it_IT:pt_BR:"
                                       "ja_JP:ko_KR:zh_CN:zh_TW:bn_IN:gu_IN:"
                                       "hi_IN:pa_IN:ta_IN:ru_RU\n");
        i18n_file_length = strlen (i18n_file_contents);
  }
  i = 0;
  line = g_string_new ("");
  for (p = i18n_file_contents; i < i18n_file_length; i += q - p, p = q) {
        gunichar character;

        character = g_utf8_get_char (p);
        q = g_utf8_next_char (p);

        /* We have a full line, let's parse it.
         *
         * It should be of the format 
         *
         * SUPPORTED=en_US.utf8:es_ES.utf8
         */
        if (character == '\n' || character == '\0') {
              gchar **installed_languages;
              int j;

              if (!g_str_has_prefix (line->str, "SUPPORTED="))
                {
                  g_string_set_size (line, 0);
                  continue;
                }
              should_fallback = FALSE;
              q = line->str + strlen ("SUPPORTED=");
              installed_languages = g_strsplit_set (q, ":\'\"", 0);

              for (j = 0; installed_languages[j] != NULL; j++) {
                    locale_without_encoding = 
                        get_less_specific_locale (installed_languages[j],
                                                  LOCALE_UP_TO_COUNTRY);
                    g_ptr_array_add (languages, locale_without_encoding);
              }
              g_strfreev (installed_languages);
              break;
        } else
          g_string_append_unichar (line, character);

        if (character == '\0')
          break;
  }

  g_string_free (line, TRUE);
  if (should_fallback) goto fallback;
  g_ptr_array_add (languages, NULL);

  preferred_languages = 
      (const char * const *) g_ptr_array_free (languages, FALSE);

  return preferred_languages;
}

static gboolean
locale_is_preferred (const char *locale)
{
  const gchar * const *preferred_languages;
  int i;
  gboolean is_preferred;
  char *language_without_encoding;
  
  preferred_languages = get_preferred_languages ();
  is_preferred = FALSE;
  for (i = 0; preferred_languages[i] != NULL; i++) {
        if (strchr (preferred_languages[i], '_') == NULL)
          continue;

        language_without_encoding = 
            get_less_specific_locale (locale, LOCALE_UP_TO_COUNTRY);

	if (strcmp (language_without_encoding, preferred_languages[i]) == 0) {
	      is_preferred = TRUE;
              g_free (language_without_encoding);
	      break;
	} 

        g_free (language_without_encoding);
  }

  return is_preferred;
}

static gboolean
model_row_is_visible (GtkTreeModel *model, GtkTreeIter *iter)
{
  gchar *locale;
  gboolean is_visible = TRUE, is_preferred;

  gtk_tree_model_get (model, iter, 
                      LOCALE_COLUMN, &locale, 
                      IS_PREFERRED_COLUMN, &is_preferred, 
                      -1);

  if ((locale != NULL) && (strcmp (locale, "-") == 0))
    {
      is_visible = FALSE;
      goto out;
    }

  if (show_all_languages)
    {
      is_visible = TRUE;
      goto out;
    }

  is_visible = is_preferred;

out:
  g_free (locale);
  return is_visible;
}
 
static void
greeter_langauge_initialize_model (void)
{
  GList *list, *li;
  GtkTreeIter iter;
  GtkTreeIter separator_iter;

  list = gdm_lang_read_locale_file (gdm_config_get_string (GDM_KEY_LOCALE_FILE));

  lang_model = gtk_list_store_new (NUM_COLUMNS,
				   G_TYPE_STRING,
				   G_TYPE_STRING,
                                   G_TYPE_STRING,
                                   G_TYPE_BOOLEAN);

  filter_model = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (GTK_TREE_MODEL (lang_model), NULL));
  gtk_tree_model_filter_set_visible_func (filter_model,
                                          (GtkTreeModelFilterVisibleFunc) 
                                          model_row_is_visible, NULL, NULL);
 
  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
		      TRANSLATED_NAME_COLUMN, _("Last Language"),
		      UNTRANSLATED_NAME_COLUMN, NULL,
		      LOCALE_COLUMN, LAST_LANGUAGE,
                      IS_PREFERRED_COLUMN, TRUE,
		      -1);

  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
		      TRANSLATED_NAME_COLUMN, _("System Default"),
		      UNTRANSLATED_NAME_COLUMN, NULL,
		      LOCALE_COLUMN, DEFAULT_LANGUAGE,
                      IS_PREFERRED_COLUMN, TRUE,
		      -1);

  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
                      TRANSLATED_NAME_COLUMN, NULL,
                      UNTRANSLATED_NAME_COLUMN, NULL,
                      LOCALE_COLUMN, "-",
                      -1);
  separator_iter = iter;

  for (li = list; li != NULL; li = li->next)
    {
      char *lang = li->data;
      char *name;
      char *untranslated;
      gboolean is_preferred;

      li->data = NULL;

      if (!locale_is_displayable (lang))
          continue;

      name = gdm_lang_name (lang,
			    FALSE /* never_encoding */,
			    TRUE /* no_group */,
			    FALSE /* untranslated */,
			    FALSE /* markup */);

      untranslated = gdm_lang_untranslated_name (lang,
						 TRUE /* markup */);
      if (untranslated == NULL)
        untranslated = g_strdup (name);
      else if (untranslated[0] == '\0')
        {
          g_free (untranslated);
          untranslated = g_strdup (name);
        }

      is_preferred = locale_is_preferred (lang);
      if (is_preferred)
        gtk_list_store_insert_before (lang_model, &iter, &separator_iter);
      else
      gtk_list_store_append (lang_model, &iter);

      gtk_list_store_set (lang_model, &iter,
			  TRANSLATED_NAME_COLUMN, name,
			  UNTRANSLATED_NAME_COLUMN, untranslated,
			  LOCALE_COLUMN, lang,
                          IS_PREFERRED_COLUMN, is_preferred,
			  -1);

      g_free (name);
      g_free (untranslated);
      g_free (lang);
    }
  g_list_free (list);
  gtk_tree_model_filter_refilter (filter_model);
}

gint
greeter_language_get_save_language (void)
{
  return savelang;
}

gchar *
greeter_language_get_language (const char *old_language)
{
  gchar *retval = NULL;

  /* Don't save language unless told otherwise */
  savelang = GTK_RESPONSE_NO;

  if (old_language == NULL)
    old_language = "";

  /* If a different language is selected */
  if (current_language != NULL && strcmp (current_language, LAST_LANGUAGE) != 0)
    {
      if (strcmp (current_language, DEFAULT_LANGUAGE) == 0)
	retval = g_strdup ("");
      else
        retval = g_strdup (current_language);

      /* User's saved language is not the chosen one */
      if (strcmp (old_language, retval) != 0)
	{
	  gchar *primary_message;
	  gchar *secondary_message;
	  char *current_name, *saved_name;

	  if (strcmp (current_language, DEFAULT_LANGUAGE) == 0)
	    current_name = g_strdup (_("System Default"));
	  else
	    current_name = gdm_lang_name (current_language,
					  FALSE /* never_encoding */,
					  TRUE /* no_group */,
					  TRUE /* untranslated */,
					  TRUE /* markup */);
	  if (strcmp (old_language, "") == 0)
	    saved_name = g_strdup (_("System Default"));
	  else
	    saved_name = gdm_lang_name (old_language,
					FALSE /* never_encoding */,
					TRUE /* no_group */,
					TRUE /* untranslated */,
					TRUE /* markup */);

	  primary_message = g_strdup_printf (_("Do you wish to make %s the default for future sessions?"),
	                                     current_name);
 	  secondary_message = g_strdup_printf (_("You have chosen %s for this session, but your default setting is "
	                                         "%s."), current_name, saved_name);
	  g_free (current_name);
	  g_free (saved_name);

	  savelang = gdm_wm_query_dialog (primary_message, secondary_message,
		_("Make _Default"), _("Just For _This Session"), TRUE);
	  g_free (primary_message);
	  g_free (secondary_message);
	}
    }
  else
    {
      retval = g_strdup (old_language);
    }

  return retval;
}

static void
selection_changed (GtkTreeSelection *selection,
		   gpointer          data)
{
  GtkTreeIter iter;

  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    {
      g_free (dialog_selected_language);
      gtk_tree_model_get (GTK_TREE_MODEL (filter_model), &iter, LOCALE_COLUMN, &dialog_selected_language, -1);
    }
}

static void
tree_row_activated (GtkTreeView         *view,
                    GtkTreePath         *path,
                    GtkTreeViewColumn   *column,
                    gpointer            data)
{
  GtkTreeIter iter;
  if (gtk_tree_model_get_iter (GTK_TREE_MODEL (filter_model), &iter, path))
    {
      g_free (dialog_selected_language);
      gtk_tree_model_get (GTK_TREE_MODEL (filter_model), &iter,
			  LOCALE_COLUMN, &dialog_selected_language,
			  -1);
      gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
  }
}

static void
more_langs_button_toggled (GtkToggleButton *expander)
{
  GtkWidget *image;
  gboolean is_active;

  if (GTK_BIN (expander)->child != NULL)
    gtk_widget_destroy (GTK_BIN (expander)->child);

  is_active = gtk_toggle_button_get_active (expander);

  image = gtk_image_new_from_stock (is_active? 
                                    GTK_STOCK_GO_DOWN : GTK_STOCK_GO_UP,
                                    GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image);
  gtk_container_add (GTK_CONTAINER (expander), image);

  show_all_languages = is_active;

  if (filter_model != NULL)
    gtk_tree_model_filter_refilter (filter_model);
}

static void 
lang_view_size_request (GtkWidget     *widget,
                        GtkRequisition *requisition)
{
  GtkWidget *dialog;
  int desired_width, desired_height;

  if (!GTK_WIDGET_MAPPED (widget))
    return;

  dialog = gtk_widget_get_toplevel (widget);

  desired_width = MIN (dialog->allocation.width - widget->allocation.width + requisition->width, 
                       (int) (.8 * gdm_wm_screen.width));
  desired_height = MIN (dialog->allocation.height - widget->allocation.height + requisition->height, 
                       (int) (.8 * gdm_wm_screen.height));

  if ((dialog->allocation.width == desired_width) && 
      (dialog->allocation.height == desired_height))
    return;

  gtk_window_resize (GTK_WINDOW (dialog), desired_width, desired_height);
}


/*
 * The button with this handler appears in the F10 menu, so it
 * cannot depend on callback data being passed in.
 */
void
greeter_language_handler (GreeterItemInfo *info, gpointer user_data)
{
  GreeterItemInfo *entry_info = greeter_lookup_id ("user-pw-entry");
  GtkWidget *entry = GNOME_CANVAS_WIDGET (entry_info->item)->widget;
  GtkWidget *expander;

  GtkWidget *view = NULL;

  if (dialog == NULL)
    {
      GtkWidget *main_vbox;
      GtkWidget *button;
      GtkWidget **tmp_p;
      GtkWidget *swindow;
      GtkWidget *label;
      char *s;

      dialog = gtk_dialog_new_with_buttons (_("Select a Language"),
#ifdef TODO
					    GTK_WINDOW (parent_window),
#endif
					    NULL,
					    0,
					    GTK_STOCK_CANCEL,
					    GTK_RESPONSE_CANCEL,
					    NULL);
					    
      button = gtk_button_new_with_mnemonic (_("Change _Language"));
      GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
      gtk_widget_show (button);
      gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button,
                                    GTK_RESPONSE_OK);
					    
      gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER_ALWAYS);
      gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
      gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);

      main_vbox = gtk_vbox_new (FALSE, 6);
      gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 5);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
                          main_vbox, TRUE, TRUE, 0);
  
      gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				       GTK_RESPONSE_OK);
      /* evil gcc warnings */
      tmp_p = &dialog;
      g_object_add_weak_pointer (G_OBJECT (dialog), (gpointer *)tmp_p);
      s = g_strdup (_("_Select the language for your session to use:"));
      label = gtk_label_new_with_mnemonic (s);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      g_free (s);
      gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
      gtk_box_pack_start (GTK_BOX (main_vbox),
			  label, FALSE, FALSE, 6);
      view = gtk_tree_view_new ();
      g_signal_connect_after (G_OBJECT (view), "size-request",
                              G_CALLBACK (lang_view_size_request), NULL);
      gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (view), TRUE);
      gtk_label_set_mnemonic_widget (GTK_LABEL (label), view);
      /* FIXME: we should handle this better, but things really look
       * bad if we aren't always LTR */
      gtk_widget_set_direction (view, GTK_TEXT_DIR_LTR);
      gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (view), FALSE);
      gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       GTK_DIALOG_MODAL,
					       NULL,
					       gtk_cell_renderer_text_new (),
					       "text", TRANSLATED_NAME_COLUMN,
					       NULL);
      gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       GTK_DIALOG_MODAL,
					       NULL,
					       gtk_cell_renderer_text_new (),
					       "markup",
					       UNTRANSLATED_NAME_COLUMN,
					       NULL);
      swindow = gtk_scrolled_window_new (NULL, NULL);
      gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swindow), GTK_SHADOW_IN);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
      gtk_container_add (GTK_CONTAINER (swindow), view);
      gtk_box_pack_start (GTK_BOX (main_vbox),
			  swindow, TRUE, TRUE, 0);
#if 0
      gtk_window_set_default_size (GTK_WINDOW (dialog),
				   MIN (400, gdm_wm_screen.width),
				   MIN (600, gdm_wm_screen.height));
#endif
      g_signal_connect (G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (view))),
			"changed",
			(GCallback) selection_changed,
			NULL);
      g_signal_connect (G_OBJECT (view),
                        "row_activated",
                        (GCallback) tree_row_activated,
                        NULL);

      expander = gtk_toggle_button_new ();

      g_signal_connect (G_OBJECT (expander),
                        "toggled", (GCallback) more_langs_button_toggled,
                        NULL);

      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (expander)) !=
          show_all_languages)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (expander),
                                      show_all_languages);
      else
        more_langs_button_toggled (GTK_TOGGLE_BUTTON (expander));

      gtk_widget_show (expander);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
        		  expander, FALSE, FALSE, 0);

      gtk_widget_show_all (dialog);
      gdm_wm_center_window (GTK_WINDOW (dialog));
    }
  gdm_wm_no_login_focus_push ();
  if (view != NULL)
    {
      GtkTreeSelection *selection;
	  
      gtk_widget_show_now (dialog);
      greeter_langauge_initialize_model ();
      gtk_tree_view_set_model (GTK_TREE_VIEW (view),
			       GTK_TREE_MODEL (filter_model));
      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
      if (selection != NULL)
	gtk_tree_selection_select_path (selection, gtk_tree_path_new_first ());
    }
  switch (gtk_dialog_run (GTK_DIALOG (dialog)))
    {
    case GTK_RESPONSE_OK:
      if (dialog_selected_language)
	current_language = g_strdup (dialog_selected_language);
      break;
    case GTK_RESPONSE_CANCEL:
    default:
      break; }

  gdm_wm_no_login_focus_pop ();

  if (dialog)
    gtk_widget_hide (dialog);
}

